"""Chat with local and remote LLMs in the context of the wordslab-notebooks environment"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_chat.ipynb.

# %% auto 0
__all__ = ['ChatTurn', 'ChatTurns', 'refresh_notebook_display', 'get_tools_schemas_and_functions', 'ToolExecutionError', 'Tools',
           'Images', 'ModelClient', 'OllamaModelClient', 'OpenRouterModelClient']

# %% ../nbs/02_chat.ipynb 2
from abc import ABC, abstractmethod
from collections.abc import Sequence as SequenceType
from pydantic import BaseModel, ConfigDict
from typing import Callable, Optional, Union, Sequence, Mapping, Any, Literal, Type

from base64 import b64decode, b64encode
from datetime import datetime
import inspect, json, re, time, traceback
from pathlib import Path
from urllib.parse import urlparse
from urllib.request import urlopen

from toolslm.funccall import get_schema, call_func
from IPython.display import display, Markdown, clear_output

from ollama import Client
from ollama._types import Options
from openai import OpenAI

from .env import WordslabEnv

# %% ../nbs/02_chat.ipynb 4
class ChatTurn:
    def __init__(self, refresh_display: Callable = None):
        self.thinking_chunks = []
        self.content_chunks = []
        self.tool_calls = {}
        self.refresh_display = refresh_display

    def append_thinking(self, chunk:str):
        self.thinking_chunks.append(chunk)
        if self.refresh_display:
            self.refresh_display()

    @property
    def thinking(self):
        return "".join(self.thinking_chunks)
    
    def append_content(self, chunk:str):
        self.content_chunks.append(chunk)
        if self.refresh_display:
            self.refresh_display()

    @property
    def content(self):
        return "".join(self.content_chunks)
    
    def append_tool_call(self, tool_name:str, params:dict):
        self.tool_calls[tool_name] = { "params": params }
        if self.refresh_display:
            self.refresh_display()

    def start_tool_call(self, tool_name:str):
        self.tool_calls[tool_name]["start_time"] = time.time()
        if self.refresh_display:
            self.refresh_display()

    def end_tool_call(self, tool_name:str, result: object):
        self.tool_calls[tool_name]["end_time"] = time.time()
        self.tool_calls[tool_name]["result"] = str(result)
        if self.refresh_display:
            self.refresh_display()

    def to_markdown(self, hide_thinking:bool=True, hide_tool_calls:bool=True):
        output = ""
        if len(self.thinking_chunks) > 0:
            if not hide_thinking:
                output += "> [Thinking]\n\n"
                output += "\n".join(f"> {line}" for line in "".join(self.thinking_chunks).splitlines()) + "\n\n"
            else:
                output += f"> [Thinking] ... thought in {sum(s.count(' ') + s.count('\n') for s in self.thinking_chunks)} words\n\n"
        if len(self.content_chunks) > 0:
            output += "".join(self.content_chunks) + "\n\n"
        if len(self.tool_calls.keys()) > 0:
            if not hide_tool_calls:
                for tool_name in self.tool_calls.keys():
                    output += "> [Tool call]\n"
                    tool_call = self.tool_calls[tool_name]
                    if "params" in tool_call:
                        output += f"> - model wants to call `{tool_name}` with parameters `{tool_call["params"]}`\n"
                    if "start_time" in tool_call:
                        output += f"> - agent called {tool_name} at {datetime.fromtimestamp(tool_call["start_time"]).strftime("%H:%M:%S")}\n"
                    if "end_time" in tool_call:
                        result = tool_call["result"]
                        output += f"> - {tool_name} returned `{result if len(result)<=100 else result[:97]+'...'}` in {(tool_call["end_time"]-tool_call["start_time"]):.3f} sec\n"
                    output += "\n"
            else:
                for tool_name in self.tool_calls.keys():
                    tool_call = self.tool_calls[tool_name]
                    if "end_time" in tool_call:
                        result = self.tool_calls[tool_name]["result"]
                        output += f"> [Tool call] ... `{tool_name}` returned `{result if len(result)<=50 else result[:47]+'...'}`\n\n"
                    elif "start_time" in tool_call:                        
                        output += f"> [Tool call] ... agent is calling `{tool_name}`\n\n"
                    elif "params" in tool_call:                        
                        output += f"> [Tool call] ... model wants to call `{tool_name}`\n\n"
        return output
 

class ChatTurns:
    def __init__(self, refresh_display:Callable = None):
        self.chat_turns = []
        self.refresh_display = refresh_display(self.chat_turns)

    def new_turn(self):
        new_turn = ChatTurn(self.refresh_display)
        self.chat_turns.append(new_turn)
        return new_turn

    
def refresh_notebook_display(hide_thinking:bool=True, hide_tool_calls:bool=True):
    def _refresh_turns(chat_turns: ChatTurns):
        def _refresh():
            clear_output(wait=True)
            output = ""
            for turn in chat_turns:
                output += turn.to_markdown(hide_thinking, hide_tool_calls)
            display(Markdown(output))
        return _refresh
    return _refresh_turns

# %% ../nbs/02_chat.ipynb 14
def _get_function_schema(func: Callable, responsesAPIFormat: bool = False):
    "Get a json schema for a python function defined with comments for all parameters"
    if responsesAPIFormat:        
        return {'type': 'function', **get_schema(func, pname='parameters')}
    else:
        return {'type': 'function', 'function': get_schema(func, pname='parameters')}

def get_tools_schemas_and_functions(funcs: Sequence[Callable], responsesAPIFormat: bool = False):
    """Get a dictionary of json schemas and callable functions which can be used for native tool calling."""
    return {func.__name__: (_get_function_schema(func, responsesAPIFormat), func) for func in funcs}

# %% ../nbs/02_chat.ipynb 17
class ToolExecutionError(Exception):
    """Raised when a tool cannot be executed safely."""

class Tools:
    """"Execute tools implemented as python functions with Large Language Models.
    The python functions must be fully documented:
    - type annotations are mandatory on all parameters and on the return type
    - a docstring after the function definition is mandatory
    - a descriptive comment after each parameter and the return type is also mandatory
    - the expected format is: one parameter by line, a traditional python comment at the end of the line
    """   
    def __init__(self, python_functions:Sequence[Callable], responsesAPIFormat:bool=False):
        self.schemas_and_functions = get_tools_schemas_and_functions(python_functions, responsesAPIFormat=responsesAPIFormat)

    def has_tool(self, tool_name:str):
        return tool_name in self.schemas_and_functions
    
    def get_schemas(self):
        return [t[0] for t in self.schemas_and_functions.values()]

    def get_schema(self, tool_name:str):
        return self.schemas_and_functions[tool_name][0]

    def get_functions(self):
        return [t[1] for t in self.schemas_and_functions.values()]

    def get_function(self, tool_name:str):
        return self.schemas_and_functions[tool_name][1]

    def call(self, tool_name:str, tool_arguments_dict:Mapping[str,Any]):
        # 1. Resolve the tool safely
        try:
            self.get_function(tool_name)
        except Exception as e:
            raise ToolExecutionError(f"Tool '{tool_name}' does not exist or could not be resolved.")
            
        # 2. Execute the tool with runtime protection
        try:
            return call_func(tool_name, tool_arguments_dict, self.get_functions(), raise_on_err=True)
        except Exception as e:
            tb = traceback.format_exc()
            raise ToolExecutionError(f"Tool '{tool_name}' raised an exception: {e}")

# %% ../nbs/02_chat.ipynb 26
class Images:
    def __init__(self, image:Union[str,Path,bytes] = None):
        self.images = []
        self.urls = []
        
        if isinstance(image, bytes):
            self.add_bytes(image)
        elif isinstance(image, Path):
            self.add_file_path(image)
        elif isinstance(image, str):
            if Images.is_web_url(image):
                self.add_web_url(image)
            else:
                self.add_file_path(image)
        else:
            raise ValueError(f'Image must be a string, Path or bytes object')

    def add_bytes(self, image_bytes: bytes):
        self.images.append(b64encode(image_bytes).decode())
        self.urls.append(False)
    
    def add_file_path(self, image_path: Union[str, Path]):
        if isinstance(image_path, str):
            image_path = Path(image_path)
        if isinstance(image_path, Path):
            if image_path.is_file():
                self.add_bytes(image_path.read_bytes())
            else:
                raise ValueError(f'Invalid image file path: {image_path}')
        else:
            raise ValueError(f'Image file path must be a string or a Path object')

    def is_web_url(image_url: str):
        try:
            parsed = urlparse(image_url)
            return parsed.scheme in ("http", "https") and bool(parsed.netloc)
        except Exception:
            return False
    
    def add_web_url(self, image_url: str): 
        if Images.is_web_url(image_url):
            self.images.append(image_url)
            self.urls.append(True)
        else:
            raise ValueError(f'Invalid image web url: {image_url}')

    def get_base64_data_or_url(self):
        return zip(self.images,self.urls)

    def get_base64_data(self):
        for i in range(len(self.images)):
            if not self.urls[i]:
                yield self.images[i]
            else:
                with urlopen(self.images[i], timeout=30) as response:
                    yield b64encode(response.read()).decode()

# %% ../nbs/02_chat.ipynb 33
class ModelClient(ABC):
    def __init__(
        self,
        model: str,        
        context_size: Optional[int] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        self.model = model
        self.base_url = base_url
        self.api_key = api_key
        self.context_size = context_size

    @abstractmethod
    def __call__(
        self,
        messages: Sequence[Mapping[str, Any]] = None,
        system_prompt: str = None,
        user_prompt: str = None,
        user_images: Images = None,
        assistant_prefill: str = None,
        tools: Tools = None,             
        max_tool_calls:int = 20,
        think: Union[bool, Literal["xhigh", "high", "medium", "low", "minimal", "none"], int, None] = None,
        web_search: bool = False,
        output_model: Type[BaseModel] = None,
        max_new_tokens: Optional[int] = None,
        seed: Optional[int] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        min_p: Optional[float] = None, 
        refresh_display: Optional[Callable] = refresh_notebook_display()
    ) -> None:
        """
        Execute a model call and return the model response.
        """
        raise NotImplementedError

    def _check_call_parameters_type(self, user_images, tools, output_model):
        if user_images and not isinstance(user_images, Images):
            raise TypeError("Argument user_images must be of type wordslab_notebooks_lib.chat.Images. Create an images object with the syntax: Images('file.jpg') or Images('https://example.com/file.jpg') or Images(image_bytes).")
        if tools and not isinstance(tools, Tools):
            raise TypeError("Argument tools must be of type wordslab_notebooks_lib.chat.Tools. Create a tools object with the syntax: Tools([func1, func2, func3]), where the parameters are documented python functions.")
        if output_model and not (isinstance(output_model, type) and issubclass(output_model, BaseModel)):
            raise TypeError("Argument output_model must be a pydantic model - a subclass of pydantic.BaseModel.")

    def _manage_messages(self, messages, system_prompt, user_prompt, images_value, assistant_prefill):
        if system_prompt:
            system_message = {'role': 'system', 'content': system_prompt}
            if len(messages) == 0:                
                messages.append(system_message) 
            elif messages[0]['role'] == 'system':
                messages[0] = system_message
            else:
                messages.insert(0, system_message)
        if user_prompt or images_value:
            ollama_format = False
            if images_value:
                ollama_format = isinstance(images_value[0], str)
            user_message = {'role': 'user'}  
            if (not images_value or ollama_format):
                if user_prompt:
                    user_message['content'] = user_prompt
                if images_value:
                    user_message['images'] = images_value
            else:
                content_list = []
                if user_prompt:  
                    content_list.append({"type": "text", "text": user_prompt})
                for image_data,is_url in images_value:
                    if is_url:
                        content_list.append({"type": "image_url", "image_url": { "url": image_data }})
                    else:
                        content_list.append({"type": "image_url", "image_url": { "url":f"data:image/jpeg;base64,{image_data}" }})
                user_message['content'] = content_list
            messages.append(user_message)
        if assistant_prefill:
            messages.append({'role': 'assistant', 'content': assistant_prefill}) 

# %% ../nbs/02_chat.ipynb 35
_whitespace_pattern = re.compile(r"\s+")

def _messages_words(messages):
    return sum([len(_whitespace_pattern.findall(message["content"][0]["text"])) if isinstance(message["content"], list) else len(_whitespace_pattern.findall(message["content"])) for message in messages if message["role"] in {"system", "user", "assistant", "tool"}])

class OllamaModelClient(ModelClient):
    def __init__(
        self,
        model: str,
        context_size: int = 32768, # This is the default value for the ollama server in wordslab-notebooks
        base_url: str = "http://localhost:11434",
        api_key: Optional[str] = None,  # If not provided, the optional key will be pulled from WordslabEnv
    ):
        super().__init__(model, context_size, base_url, api_key)

        # Initialize API client
        if not api_key:
            env = WordslabEnv()
            api_key = env.cloud_ollama_api_key
        if api_key:
            self.has_api_key = True
            headers = {'Authorization': 'Bearer ' + api_key}
        else:
            self.has_api_key = False
            headers = {}
        self.client = Client(host=self.base_url, headers=headers)
        
        # Load model in memory with the right context length
        print(f"ollama: loading model {self.model} with context size {self.context_size} ... ", end="");
        self.client.chat(model=self.model, messages=[{'role': 'user', 'content': 'say yes'}], options=Options(num_ctx=self.context_size, num_predict=1))
        print(f"ok");

    def __call__(
        self,
        messages: Sequence[Mapping[str, Any]] = None,
        system_prompt: str = None,
        user_prompt: str = None,
        user_images: Images = None,
        assistant_prefill: str = None,
        tools: Tools = None,             
        max_tool_calls:int = 20,
        think: Union[bool, Literal["low", "medium", "high"], None] = None,
        web_search: bool = False,
        output_model: Type[BaseModel] = None,
        max_new_tokens: Optional[int] = None,
        seed: Optional[int] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        min_p: Optional[float] = None, 
        refresh_display: Optional[Callable] = refresh_notebook_display()
    ) -> None:
        # Reset response field
        self.response = None
        
        # Check parameters type
        super()._check_call_parameters_type(user_images, tools, output_model)

        # Images
        images_value = None
        if user_images:
            images_value = list(user_images.get_base64_data())

        # Messages
        if messages is None:
            messages = []
        super()._manage_messages(messages, system_prompt, user_prompt, images_value, assistant_prefill)        

        # Web search
        if web_search:
            if self.has_api_key:
                tools = Tools([self.client.web_search, self.client.web_fetch] + (tools if tools else []))
            else:
                raise RuntimeError("Web search requires an api_key in OllamaModelClient constructor.")
                
        chat_turns = ChatTurns(refresh_display)
        
        for i in range(max_tool_calls):
            # Immediate user feedback
            print(f"ollama: processing {_messages_words(messages)} words with `{self.model}` ...")
            
            # Observable conversation turn
            chat_turn = chat_turns.new_turn()
            if assistant_prefill and len(chat_turns.chat_turns) == 1:
                chat_turn.append_content(assistant_prefill)
            
            stream = self.client.chat(
                model = self.model,
                messages = messages,
                tools = tools.get_schemas() if tools else None,
                stream = True,
                think = think,
                format = output_model.model_json_schema() if output_model else None,
                options = Options(num_ctx = self.context_size, num_predict = max_new_tokens, seed = seed,
                                  temperature = temperature, top_k=top_k, top_p=top_p, min_p=min_p)
            )
        
            # Streaming: accumulate the partial fields
            tool_calls = []        
            for chunk in stream:
                if chunk.message.thinking:
                    chat_turn.append_thinking(chunk.message.thinking)                
                if chunk.message.content:
                    chat_turn.append_content(chunk.message.content)
                if chunk.message.tool_calls:
                    tool_calls.extend(chunk.message.tool_calls)
                    for tc in chunk.message.tool_calls:
                        chat_turn.append_tool_call(tc.function.name, tc.function.arguments)
            
            # append accumulated fields to the messages
            if chat_turn.thinking or chat_turn.content or tool_calls:
                messages.append({'role': 'assistant', 'thinking': chat_turn.thinking, 'content': chat_turn.content, 'tool_calls': tool_calls})
        
            # end the loop if there is no more tool calls
            if not tool_calls: 
                # Set response field
                if 'content' in messages[-1]:
                    self.response = messages[-1]['content']
                if output_model:
                    self.response = output_model.model_validate_json(self.response)
                return 
                
            # execute tool calls  
            else:    
                for tc in tool_calls:
                    if tools.has_tool(tc.function.name):
                        chat_turn.start_tool_call(tc.function.name)
                        result = tools.call(tc.function.name, tc.function.arguments)
                        chat_turn.end_tool_call(tc.function.name, result)
                    else:
                        result = 'Unknown tool'
            
                    # append tool call result to the messages 
                    messages.append({'role': 'tool', 'tool_name': tc.function.name, 'content': str(result)})
    
            # continue the loop after tool calls

# %% ../nbs/02_chat.ipynb 63
class OpenRouterModelClient(ModelClient):
    def __init__(
        self,
        model: str,
        context_size: Optional[int] = None, # For OpenRouter this parameter is ignored, we inherit the remote model config
        base_url: str = "https://openrouter.ai/api/v1",
        api_key: Optional[str] = None, # If not provided, the mandatory key will be pulled from WordslabEnv
    ):
        super().__init__(model, context_size, base_url, api_key)

        # Initialize API client
        if not api_key:
            env = WordslabEnv()
            api_key = env.cloud_openrouter_api_key
        self.client = OpenAI(base_url=base_url, api_key=api_key)
        
        # Check connection
        print(f"openrouter: testing model {self.model} ... ", end="");
        self.client.chat.completions.create(model=self.model, messages=[{'role': 'user', 'content': 'say yes'}], max_tokens=16)
        print(f"ok");

    def __call__(
        self,
        messages: Sequence[Mapping[str, Any]] = None,
        system_prompt: str = None,
        user_prompt: str = None,
        user_images: Images = None,
        assistant_prefill: str = None,
        tools: Tools = None,             
        max_tool_calls:int = 20,
        think: Union[bool, Literal["xhigh", "high", "medium", "low", "minimal", "none"], int, None] = None,
        web_search: bool = False,
        output_model: Type[BaseModel] = None,
        max_new_tokens: Optional[int] = None,
        seed: Optional[int] = None,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        min_p: Optional[float] = None, 
        refresh_display: Optional[Callable] = refresh_notebook_display()
    ) -> None:
        # Reset response field
        self.response = None
        self.annotations = None
        
        # Check parameters type
        super()._check_call_parameters_type(user_images, tools, output_model)

        # Images
        images_value = None
        if user_images:
            images_value = list(user_images.get_base64_data_or_url())
        
        # Messages
        if messages is None:
            messages = []
        super()._manage_messages(messages, system_prompt, user_prompt, images_value, assistant_prefill)

        # Map "think" → reasoning > enabled / reasoning > effort / resoning > max_tokens
        reasoning = None
        if think is True:
            reasoning = {"reasoning": {"enabled": True}} # reasoning on/off
        elif think in ("low", "medium", "high"):
            reasoning = {"reasoning": {"effort": think}} # "xhigh", "high", "medium", "low", "minimal" or "none" (OpenAI-style)
        elif isinstance(think, int):
            reasoning = {"reasoning": {"max_tokens": think}} # specified token budget for extended thinking (Anthropic-style)

        # Structured outputs
        response_format = None
        if output_model:
            schema = output_model.model_json_schema()
            schema["additionalProperties"] = False
            response_format = {
                    "type": "json_schema",
                    "json_schema": {
                        "name": output_model.__name__,
                        "schema": schema,
                        "strict": True,
                    },
                }
        
        chat_turns = ChatTurns(refresh_display)
        
        for i in range(max_tool_calls):
            # Immediate user feedback
            print(f"openrouter: processing {_messages_words(messages)} words with `{self.model + (":online" if web_search else "")}` ...")
            if web_search:
                print("-> web search activated")
            
            # Observable conversation turn
            chat_turn = chat_turns.new_turn()
            if assistant_prefill and len(chat_turns.chat_turns) == 1:
                chat_turn.append_content(assistant_prefill)            
            
            stream = self.client.chat.completions.create(
                model = self.model + (":online" if web_search else ""),
                messages = messages,
                tools = tools.get_schemas() if tools else None,
                stream = True,
                extra_body = reasoning,
                response_format = response_format,
                max_tokens = max_new_tokens,
                seed = seed,
                temperature = temperature,
                top_p = top_p,
            )
        
            # Streaming: accumulate the partial fields
            tool_calls = {} 
            annotations = None
            for chunk in stream:
                delta = chunk.choices[0].delta
                if hasattr(delta, "reasoning") and delta.reasoning:
                    chat_turn.append_thinking(delta.reasoning)                
                if hasattr(delta, "content") and delta.content:
                    chat_turn.append_content(delta.content)
                # web search results - standardized by openrouter
                if hasattr(delta, "annotations") and delta.annotations:
                    annotations = delta.annotations
                if hasattr(delta, "tool_calls") and delta.tool_calls:
                    for tool_call in delta.tool_calls:
                        idx = tool_call.index
                        # First tool call chunk
                        if idx not in tool_calls:
                            tool_calls[idx] = {
                                "id": tool_call.id,  # only present in first chunk
                                "name": tool_call.function.name,
                                "arguments": ""
                            }       
                        # Append streamed argument fragments
                        tool_calls[idx]["arguments"] += (
                            tool_call.function.arguments or ""
                        )
                        
            # We need to wait the end of the stream to make sure the tool calls are complete
            for tc in tool_calls.values():
                chat_turn.append_tool_call(tc["name"], tc["arguments"])
            
            # Append accumulated fields to the messages
            if chat_turn.content or tool_calls:
                messages.append({
                    "role": "assistant",
                    "content": chat_turn.content,
                    "tool_calls": [
                        {
                            "id": tc["id"],
                            "type": "function",
                            "function": {
                                "name": tc["name"],
                                "arguments": tc["arguments"]
                            }
                        }
                        for tc in tool_calls.values()
                    ]
                })
                
            # end the loop if there is no more tool calls
            if not tool_calls: 
                # Set response field
                if 'content' in messages[-1]:
                    self.response = messages[-1]['content']
                if annotations:
                    self.annotations = annotations
                if output_model:
                    self.response = output_model.model_validate_json(self.response)
                return      
                
            # execute tool calls  
            else:    
                for tc in tool_calls.values():
                    if tools.has_tool(tc["name"]):
                        chat_turn.start_tool_call(tc["name"])
                        result = tools.call(tc["name"], json.loads(tc["arguments"]))
                        chat_turn.end_tool_call(tc["name"], result)
                    else:
                        result = 'Unknown tool'
            
                    # append tool call result to the messages 
                    messages.append({"role": "tool", "tool_call_id": tc["id"], "content": str(result)})
    
            # continue the loop after tool calls



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/wordslab-notebooks-lib',
                'doc_host': 'https://wordslab-org.github.io',
                'git_url': 'https://github.com/wordslab-org/wordslab-notebooks-lib',
                'lib_path': 'wordslab_notebooks_lib'},
  'syms': { 'wordslab_notebooks_lib.chat': { 'wordslab_notebooks_lib.chat.ChatTurn': ( 'chat.html#chatturn',
                                                                                       'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.__init__': ( 'chat.html#chatturn.__init__',
                                                                                                'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.append_content': ( 'chat.html#chatturn.append_content',
                                                                                                      'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.append_thinking': ( 'chat.html#chatturn.append_thinking',
                                                                                                       'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.append_tool_call': ( 'chat.html#chatturn.append_tool_call',
                                                                                                        'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.content': ( 'chat.html#chatturn.content',
                                                                                               'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.end_tool_call': ( 'chat.html#chatturn.end_tool_call',
                                                                                                     'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.start_tool_call': ( 'chat.html#chatturn.start_tool_call',
                                                                                                       'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.thinking': ( 'chat.html#chatturn.thinking',
                                                                                                'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurn.to_markdown': ( 'chat.html#chatturn.to_markdown',
                                                                                                   'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurns': ( 'chat.html#chatturns',
                                                                                        'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurns.__init__': ( 'chat.html#chatturns.__init__',
                                                                                                 'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ChatTurns.new_turn': ( 'chat.html#chatturns.new_turn',
                                                                                                 'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images': ('chat.html#images', 'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images.__init__': ( 'chat.html#images.__init__',
                                                                                              'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images.add_bytes': ( 'chat.html#images.add_bytes',
                                                                                               'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images.add_file_path': ( 'chat.html#images.add_file_path',
                                                                                                   'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images.add_web_url': ( 'chat.html#images.add_web_url',
                                                                                                 'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images.get_base64_data': ( 'chat.html#images.get_base64_data',
                                                                                                     'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images.get_base64_data_or_url': ( 'chat.html#images.get_base64_data_or_url',
                                                                                                            'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Images.is_web_url': ( 'chat.html#images.is_web_url',
                                                                                                'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ModelClient': ( 'chat.html#modelclient',
                                                                                          'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ModelClient.__call__': ( 'chat.html#modelclient.__call__',
                                                                                                   'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ModelClient.__init__': ( 'chat.html#modelclient.__init__',
                                                                                                   'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ModelClient._check_call_parameters_type': ( 'chat.html#modelclient._check_call_parameters_type',
                                                                                                                      'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ModelClient._manage_messages': ( 'chat.html#modelclient._manage_messages',
                                                                                                           'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.OllamaModelClient': ( 'chat.html#ollamamodelclient',
                                                                                                'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.OllamaModelClient.__call__': ( 'chat.html#ollamamodelclient.__call__',
                                                                                                         'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.OllamaModelClient.__init__': ( 'chat.html#ollamamodelclient.__init__',
                                                                                                         'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.OpenRouterModelClient': ( 'chat.html#openroutermodelclient',
                                                                                                    'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.OpenRouterModelClient.__call__': ( 'chat.html#openroutermodelclient.__call__',
                                                                                                             'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.OpenRouterModelClient.__init__': ( 'chat.html#openroutermodelclient.__init__',
                                                                                                             'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.ToolExecutionError': ( 'chat.html#toolexecutionerror',
                                                                                                 'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools': ('chat.html#tools', 'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools.__init__': ( 'chat.html#tools.__init__',
                                                                                             'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools.call': ( 'chat.html#tools.call',
                                                                                         'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools.get_function': ( 'chat.html#tools.get_function',
                                                                                                 'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools.get_functions': ( 'chat.html#tools.get_functions',
                                                                                                  'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools.get_schema': ( 'chat.html#tools.get_schema',
                                                                                               'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools.get_schemas': ( 'chat.html#tools.get_schemas',
                                                                                                'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.Tools.has_tool': ( 'chat.html#tools.has_tool',
                                                                                             'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat._get_function_schema': ( 'chat.html#_get_function_schema',
                                                                                                   'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat._messages_words': ( 'chat.html#_messages_words',
                                                                                              'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.get_tools_schemas_and_functions': ( 'chat.html#get_tools_schemas_and_functions',
                                                                                                              'wordslab_notebooks_lib/chat.py'),
                                             'wordslab_notebooks_lib.chat.refresh_notebook_display': ( 'chat.html#refresh_notebook_display',
                                                                                                       'wordslab_notebooks_lib/chat.py')},
            'wordslab_notebooks_lib.core': { 'wordslab_notebooks_lib.core.Wordslab': ( 'core.html#wordslab',
                                                                                       'wordslab_notebooks_lib/core.py'),
                                             'wordslab_notebooks_lib.core.Wordslab.__init__': ( 'core.html#wordslab.__init__',
                                                                                                'wordslab_notebooks_lib/core.py')},
            'wordslab_notebooks_lib.env': { 'wordslab_notebooks_lib.env.WordslabEnv': ( 'env.html#wordslabenv',
                                                                                        'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.__init__': ( 'env.html#wordslabenv.__init__',
                                                                                                 'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.read_user_env_var': ( 'env.html#wordslabenv.read_user_env_var',
                                                                                                          'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.setup_github': ( 'env.html#wordslabenv.setup_github',
                                                                                                     'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.setup_huggingface': ( 'env.html#wordslabenv.setup_huggingface',
                                                                                                          'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.setup_ollamacloud': ( 'env.html#wordslabenv.setup_ollamacloud',
                                                                                                          'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.setup_openrouter': ( 'env.html#wordslabenv.setup_openrouter',
                                                                                                         'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.setup_pypi': ( 'env.html#wordslabenv.setup_pypi',
                                                                                                   'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.setup_replicate': ( 'env.html#wordslabenv.setup_replicate',
                                                                                                        'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.setup_tavily': ( 'env.html#wordslabenv.setup_tavily',
                                                                                                     'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env.WordslabEnv.write_user_env_var': ( 'env.html#wordslabenv.write_user_env_var',
                                                                                                           'wordslab_notebooks_lib/env.py'),
                                            'wordslab_notebooks_lib.env._expand_with_current_env': ( 'env.html#_expand_with_current_env',
                                                                                                     'wordslab_notebooks_lib/env.py')},
            'wordslab_notebooks_lib.notebook': { 'wordslab_notebooks_lib.notebook.Inspector._get_info': ( 'notebook.html#inspector._get_info',
                                                                                                          'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.InteractiveShell.get_tools_schemas_and_functions': ( 'notebook.html#interactiveshell.get_tools_schemas_and_functions',
                                                                                                                                       'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.InteractiveShell.get_variables_values': ( 'notebook.html#interactiveshell.get_variables_values',
                                                                                                                            'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.InteractiveShell.user_items': ( 'notebook.html#interactiveshell.user_items',
                                                                                                                  'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook': ( 'notebook.html#wordslabnotebook',
                                                                                                       'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.JupyterlabExtensionComm': ( 'notebook.html#wordslabnotebook.jupyterlabextensioncomm',
                                                                                                                               'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.__init__': ( 'notebook.html#wordslabnotebook.__init__',
                                                                                                                'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook._ensure_jupyterlab_extension': ( 'notebook.html#wordslabnotebook._ensure_jupyterlab_extension',
                                                                                                                                    'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.add_cell': ( 'notebook.html#wordslabnotebook.add_cell',
                                                                                                                'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.cell_id': ( 'notebook.html#wordslabnotebook.cell_id',
                                                                                                               'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.chat': ( 'notebook.html#wordslabnotebook.chat',
                                                                                                            'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.content': ( 'notebook.html#wordslabnotebook.content',
                                                                                                               'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.delete_cell': ( 'notebook.html#wordslabnotebook.delete_cell',
                                                                                                                   'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.get_context_for_llm': ( 'notebook.html#wordslabnotebook.get_context_for_llm',
                                                                                                                           'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.get_tools_schemas_and_functions': ( 'notebook.html#wordslabnotebook.get_tools_schemas_and_functions',
                                                                                                                                       'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.get_variables_values': ( 'notebook.html#wordslabnotebook.get_variables_values',
                                                                                                                            'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.jupyterlab_extension_version': ( 'notebook.html#wordslabnotebook.jupyterlab_extension_version',
                                                                                                                                    'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.path': ( 'notebook.html#wordslabnotebook.path',
                                                                                                            'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.read_cell': ( 'notebook.html#wordslabnotebook.read_cell',
                                                                                                                 'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.run_cell': ( 'notebook.html#wordslabnotebook.run_cell',
                                                                                                                'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.set_ollama_chat_model': ( 'notebook.html#wordslabnotebook.set_ollama_chat_model',
                                                                                                                             'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.set_openrouter_chat_model': ( 'notebook.html#wordslabnotebook.set_openrouter_chat_model',
                                                                                                                                 'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.show_object_members': ( 'notebook.html#wordslabnotebook.show_object_members',
                                                                                                                           'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.show_variables_and_functions': ( 'notebook.html#wordslabnotebook.show_variables_and_functions',
                                                                                                                                    'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.WordslabNotebook.update_cell': ( 'notebook.html#wordslabnotebook.update_cell',
                                                                                                                   'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._cell_output_to_xml': ( 'notebook.html#_cell_output_to_xml',
                                                                                                          'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._cell_to_xml': ( 'notebook.html#_cell_to_xml',
                                                                                                   'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._cells_to_notebook_xml': ( 'notebook.html#_cells_to_notebook_xml',
                                                                                                             'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._find_frame_dict': ( 'notebook.html#_find_frame_dict',
                                                                                                       'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._get_schema': ( 'notebook.html#_get_schema',
                                                                                                  'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._mime_bundle_to_text': ( 'notebook.html#_mime_bundle_to_text',
                                                                                                           'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._safe_attr_doc': ( 'notebook.html#_safe_attr_doc',
                                                                                                     'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._safe_getattr': ( 'notebook.html#_safe_getattr',
                                                                                                    'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook._safe_str': ( 'notebook.html#_safe_str',
                                                                                                'wordslab_notebooks_lib/notebook.py'),
                                                 'wordslab_notebooks_lib.notebook.find_var': ( 'notebook.html#find_var',
                                                                                               'wordslab_notebooks_lib/notebook.py')},
            'wordslab_notebooks_lib.tools': { 'wordslab_notebooks_lib.tools._absolutify_imgs': ( 'tools.html#_absolutify_imgs',
                                                                                                 'wordslab_notebooks_lib/tools.py'),
                                              'wordslab_notebooks_lib.tools._convert_math': ( 'tools.html#_convert_math',
                                                                                              'wordslab_notebooks_lib/tools.py'),
                                              'wordslab_notebooks_lib.tools.create': ( 'tools.html#create',
                                                                                       'wordslab_notebooks_lib/tools.py'),
                                              'wordslab_notebooks_lib.tools.insert': ( 'tools.html#insert',
                                                                                       'wordslab_notebooks_lib/tools.py'),
                                              'wordslab_notebooks_lib.tools.read_url': ( 'tools.html#read_url',
                                                                                         'wordslab_notebooks_lib/tools.py'),
                                              'wordslab_notebooks_lib.tools.scrape_url': ( 'tools.html#scrape_url',
                                                                                           'wordslab_notebooks_lib/tools.py'),
                                              'wordslab_notebooks_lib.tools.str_replace': ( 'tools.html#str_replace',
                                                                                            'wordslab_notebooks_lib/tools.py'),
                                              'wordslab_notebooks_lib.tools.view': ('tools.html#view', 'wordslab_notebooks_lib/tools.py')}}}



"""Predefined tools for LLMs"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_tools.ipynb.

# %% auto 0
__all__ = ['view', 'create', 'insert', 'str_replace', 'scrape_url', 'read_url']

# %% ../nbs/03_tools.ipynb 2
from bs4 import BeautifulSoup
from cloudscraper import create_scraper
from html2text import HTML2Text
from pathlib import Path
from urllib.parse import urljoin, urlparse
import re
from textwrap import dedent

# %% ../nbs/03_tools.ipynb 9
def view(path:str,  # The path to the file or directory to view
         view_range:tuple[int,int]=None, # Optional array of two integers specifying the start and end line numbers to view. Line numbers are 1-indexed, and -1 for the end line means read to the end of the file. This parameter only applies when viewing files, not directories.
         nums:bool=False # Optionally prefix all lines of the file with a line number
        ) -> str:
    'Examine the contents of a file or list the contents of a directory. It can read the entire file or a specific range of lines. With or without line numbers.'
    try:
        p = Path(path).expanduser().resolve()
        if not p.exists(): return f'Error: File not found: {p}'
        if p.is_dir():
            res = [str(f) for f in p.glob('**/*') 
                   if not any(part.startswith('.') for part in f.relative_to(p).parts)]
            return f'Directory contents of {p}:\n' + '\n'.join(res)
        
        lines = p.read_text().splitlines()
        s,e = 1,len(lines)
        if view_range:
            s,e = view_range
            if not (1 <= s <= len(lines)): return f'Error: Invalid start line {s}'
            if e != -1 and not (s <= e <= len(lines)): return f'Error: Invalid end line {e}'
            lines = lines[s-1:None if e==-1 else e]
            
        return '\n'.join([f'{i+s-1:6d} │ {l}' for i,l in enumerate(lines,1)] if nums else lines)
    except Exception as e: return f'Error viewing file: {str(e)}'
     

# %% ../nbs/03_tools.ipynb 11
def create(path: str, # The path where the new file should be created
           file_text: str, # The text content to write to the new file
           overwrite:bool=False # Allows overwriting an existing file
          ) -> str:
    'Creates a new file with the given text content at the specified path'
    try:
        p = Path(path)
        if p.exists():
            if not overwrite: return f'Error: File already exists: {p}'
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(file_text)
        return f'Created file {p} containing:\n{file_text}'
    except Exception as e: return f'Error creating file: {str(e)}'     

# %% ../nbs/03_tools.ipynb 13
def insert(path: str,  # The path to the file to modify
           insert_line: int, # The line number after which to insert the text (0 for beginning of file)
           new_str: str # The text to insert
          ) -> str: 
    'Insert text at a specific line number in a file.'
    try:
        p = Path(path)
        if not p.exists(): return f'Error: File not found: {p}'
            
        content = p.read_text().splitlines()
        if not (0 <= insert_line <= len(content)): return f'Error: Invalid line number {insert_line}'
            
        content.insert(insert_line, new_str)
        new_content = '\n'.join(content)
        p.write_text(new_content)
        return f'Inserted text at line {insert_line} in {p}.\nNew contents:\n{new_content}'
    except Exception as e: return f'Error inserting text: {str(e)}'

# %% ../nbs/03_tools.ipynb 15
def str_replace(path: str, # The path to the file to modify
                old_str: str, # The text to replace (must match exactly, including whitespace and indentation)
                new_str: str # The new text to insert in place of the old text
               ) -> str:
    'Replace a specific string in a file with a new string. This is used for making precise edits.'
    try:
        p = Path(path)
        if not p.exists(): return f'Error: File not found: {p}'
            
        content = p.read_text()
        count = content.count(old_str)
        
        if count == 0: return 'Error: Text not found in file'
        if count > 1: return f'Error: Multiple matches found ({count})'
            
        new_content = content.replace(old_str, new_str, 1)
        p.write_text(new_content)
        return f'Replaced text in {p}.\nNew contents:\n{new_content}'
    except Exception as e: return f'Error replacing text: {str(e)}'

# %% ../nbs/03_tools.ipynb 23
def _absolutify_imgs(md, base_url):
    """This function rewrites Markdown image links so their URLs become absolute, using a base URL.
    - md: a Markdown string
    - base_url: the base URL used to resolve relative path
    """
    def fix(m):
        alt, img_url = m.group(1), m.group(2)
        if not img_url.startswith('http'): img_url = urljoin(base_url, img_url)
        return f'![{alt}]({img_url})'
    return re.sub(r'!\[(.*?)\]\((.*?)\)', fix, md)


# %% ../nbs/03_tools.ipynb 24
def _convert_math(soup, mode):
    """This function walks an HTML/XML document, finds MathML math elements, and replaces them with TeX/LaTeX markup suitable for text-based math renderers (like Markdown, MathJax, or KaTeX).
    - soup: a BeautifulSoup object representing parsed HTML/XML
    - mode: controls the inline math delimiter style
    Two inline math styles are supported:
    - mode='dollar': Inline math $x$ Display math $$x$$
    - other	Inline math \\(...\\) Display math $$x$$
    """
    for math in soup.find_all('math'):
        annot = math.find('annotation', {'encoding': 'application/x-tex'})
        if not annot:
            continue
        tex, display = annot.text.strip(), math.get('display') == 'block'
        if mode == 'dollar':
            wrap = f'$${tex}$$' if display else f'${tex}$'
        else:
            wrap = f'$${tex}$$' if display else f'\\({tex}\\)'
        math.replace_with(wrap)

# %% ../nbs/03_tools.ipynb 25
def scrape_url(url):
    "Get the html content of a web page using the cloudscraper library to bypass Cloudflare's anti-bot page."
    return create_scraper().get(url)

def read_url(url: str, as_md: bool = True, extract_section: bool = True, selector: str = None, math_mode: str = None):
    """This functions extracts a web page information for LLM ingestion
    1. Downloads a web page
    2. Parses HTML
    3. Optionally extracts a specific section (fragment or CSS selector)
    4. Converts MathML → LaTeX
    5. Optionally converts HTML → Markdown
    6. Convert code sections to fenced markdown blocks
    7. Makes image URLs absolute
    8. Returns the processed text
    """
    o = scrape_url(url)
    res, ctype = o.text, o.headers.get('content-type').split(';')[0]
    soup = BeautifulSoup(res, 'lxml')

    if selector:
        res = '\n\n'.join(str(s) for s in soup.select(selector))
    elif extract_section:
        parsed = urlparse(url)
        if parsed.fragment:
            section = soup.find(id=parsed.fragment)
            if section:
                elements = [section]
                current = section.next_sibling
                while current:
                    if hasattr(current, 'name') and current.name == section.name: break
                    elements.append(current)
                    current = current.next_sibling
                res = ''.join(str(el) for el in elements)
            else:
                res = ''
    else:
        res = str(soup)

    if math_mode:
        res_soup = BeautifulSoup(res, 'lxml')
        _convert_math(res_soup, math_mode)
        res = str(res_soup)

    if as_md and ctype == 'text/html':
        h = HTML2Text()
        h.body_width = 0
        # Handle code blocks
        h.mark_code = True
        res = h.handle(res)
        def _f(m): return f'```\n{dedent(m.group(1))}\n```'
        res = re.sub(r'\[code]\s*\n(.*?)\n\[/code]', _f, res or '', flags=re.DOTALL).strip()
        # Handle image urls
        res = _absolutify_imgs(res, urljoin(url, s['href'] if (s := soup.find('base')) else ''))
        # Handle math blocks
        if math_mode == 'safe':
            res = res.replace('\\\\(', '\\(').replace('\\\\)', '\\)')

    return res

# %% ../nbs/03_tools.ipynb 43
# import ...



"""Access wordslab-notebooks environment version, platform, urls, ports, directories (install / apps / data / models), and pre-installed models."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_env.ipynb.

# %% auto 0
__all__ = ['WordslabEnv']

# %% ../nbs/01_env.ipynb 2
import os, re, subprocess
from pathlib import Path
from typing import Optional

from fastcore.utils import patch

# %% ../nbs/01_env.ipynb 5
class WordslabEnv:
    def __init__(self):
        # wordslab-notebooks version and platform
        self.version = os.environ["WORDSLAB_VERSION"]
        self.platform = os.environ["WORDSLAB_PLATFORM"]

        # wordslab-notebooks external urls
        self.url_dashboard = os.environ["DASHBOARD_URL"]
        self.url_openwebui = os.environ["OPENWEBUI_URL"]
        self.url_jupyterlab = os.environ["JUPYTERLAB_URL"]
        self.url_vscode = os.environ["VSCODE_URL"]
        self.url_userapp1 = os.environ["USER_APP1_URL"]
        self.url_userapp2 = os.environ["USER_APP2_URL"]
        self.url_userapp3 = os.environ["USER_APP3_URL"]
        self.url_userapp4 = os.environ["USER_APP4_URL"]
        self.url_userapp5 = os.environ["USER_APP5_URL"]

        # wordslab-notebooks internal ports
        self.port_ollama = 11434
        self.port_vllm = 8000
        self.port_docling = 5001   
        self.port_userapp1 = os.environ["USER_APP1_PORT"]
        self.port_userapp2 = os.environ["USER_APP2_PORT"]
        self.port_userapp3 = os.environ["USER_APP3_PORT"]
        self.port_userapp4 = os.environ["USER_APP4_PORT"]
        self.port_userapp5 = os.environ["USER_APP5_PORT"]
        
        # wordslab-notebooks install directories
        self.dir_home = os.environ["WORDSLAB_HOME"]
        self.dir_scripts = os.environ["WORDSLAB_SCRIPTS"]     
        self.dir_python = os.environ["UV_PYTHON_INSTALL_DIR"]  
        self.dir_workspace = os.environ["WORDSLAB_WORKSPACE"]
        self.dir_models = os.environ["WORDSLAB_MODELS"]

        # wordslab-notebooks applications install directories
        self.dir_openwebui = os.environ["OPENWEBUI_ENV"]
        self.dir_jupyterlab = os.environ["JUPYTERLAB_ENV"]
        self.dir_vscode = os.environ["VSCODE_DIR"]
        self.dir_ollama = os.environ["OLLAMA_DIR"]
        self.dir_docling = os.environ["DOCLING_ENV"]  

        # wordslab-notebooks applications data directories
        self.dir_openwebui_data = os.environ["OPENWEBUI_DATA"]
        self.dir_jupyterlab_data = os.environ["JUPYTER_DATA"]
        self.dir_vscode_data = os.environ["VSCODE_DATA"]
        self.dir_docling_data = os.environ["DOCLING_DATA"]
        
        # wordslab-notebooks models directories
        self.dir_models_ollama = os.environ["OLLAMA_MODELS"] 
        self.dir_models_vllm = os.environ["HF_HOME"]
        self.dir_models_hugginface = os.environ["HF_HOME"]
        self.dir_models_fastai = os.environ["FASTAI_HOME"] 
        self.dir_models_pytorch = os.environ["TORCH_HOME"] 
        self.dir_models_keras = os.environ["KERAS_HOME"]
        self.dir_models_tensorflow = os.environ["TFHUB_CACHE_DIR"]         
        self.dir_models_docling = os.environ["DOCLING_MODELS"]
        
        # wordslab-notebooks default models
        self.default_model_chat = os.environ["OLLAMA_CHAT_MODEL"]
        self.default_model_code = os.environ["OLLAMA_CODE_MODEL"]
        self.default_model_agent = os.environ["OLLAMA_AGENT_MODEL"]        
        self.default_model_embedding = os.environ["OLLAMA_EMBED_MODEL"]
        self.default_model_context_length = os.environ["OLLAMA_CONTEXT_LENGTH"]

        # external cloud services - api keys
        # Note: you need to reload the WordslabEnv() object after you define one of these environment variables
        self.cloud_openrouter_api_key_var = "OPENROUTER_API_KEY"
        self.cloud_openrouter_api_key = os.environ.get(self.cloud_openrouter_api_key_var)
        self.cloud_replicate_api_token_var = "REPLICATE_API_TOKEN"
        self.cloud_replicate_api_token = os.environ.get(self.cloud_replicate_api_token_var)
        self.cloud_tavily_api_key_var = "TAVILY_API_KEY"
        self.cloud_tavily_api_key = os.environ.get(self.cloud_tavily_api_key_var)
        self.cloud_ollama_api_key_var = "OLLAMA_API_KEY"
        self.cloud_ollama_api_key = os.environ.get(self.cloud_ollama_api_key_var)
        self.cloud_huggingface_access_token_var = "HF_TOKEN"
        self.cloud_huggingface_access_token = os.environ.get(self.cloud_huggingface_access_token_var)

# %% ../nbs/01_env.ipynb 63
_ENV_REF_PATTERN = re.compile(
    r"""
    \$(\w+)|           # $VAR
    \${([^}]+)}        # ${VAR}
    """,
    re.VERBOSE,
)

def _expand_with_current_env(value: str) -> str:
    """
    Expand $VAR and ${VAR} using os.environ.
    Unknown variables are left unchanged.
    """

    def replacer(match: re.Match) -> str:
        var_name = match.group(1) or match.group(2)
        return os.environ.get(var_name, match.group(0))

    return _ENV_REF_PATTERN.sub(replacer, value)

@patch
def write_user_env_var(self: WordslabEnv, var_name: str, var_value: str) -> None:
    """
    Write or update an environment variable in
    $WORDSLAB_WORKSPACE/.secrets/user-env.bashrc

    The file is created if it doesn't exist.
    The variable is created if it doesn't exist and updated if it already exist.
    The value is safely quoted for Bash while preserving $VAR references.

    The new variable value is automatically injected in the current python process.
    The environment variable will be automatically loaded each time you start a new shell.
    But you need to restart all the currently running shells and other processes to see the new variable or value.
    """

    if not re.fullmatch(r"[A-Za-z_][A-Za-z0-9_]*", var_name):
        raise ValueError(f"Invalid environment variable name: {var_name}")

    workspace = os.environ.get("WORDSLAB_WORKSPACE")
    if not workspace:
        raise EnvironmentError("WORDSLAB_WORKSPACE is not set")

    env_file = Path(workspace) / ".secrets" / "user-env.bashrc"
    env_file.parent.mkdir(parents=True, exist_ok=True)

    # Escape value for double-quoted Bash string (but keep $ unescaped)
    escaped_value = (
        var_value
        .replace("\\", "\\\\")
        .replace('"', '\\"')
        .replace("`", "\\`")
    )

    new_line = f'export {var_name}="{escaped_value}"'

    if env_file.exists():
        lines = env_file.read_text().splitlines()
    else:
        lines = []

    var_pattern = re.compile(rf"^\s*(?:export\s+)?{re.escape(var_name)}\s*=")
    updated = False

    for i, line in enumerate(lines):
        if var_pattern.match(line):
            lines[i] = new_line
            updated = True
            break

    if not updated:
        if lines and lines[-1].strip() != "":
            lines.append("")
        lines.append(new_line)

    env_file.write_text("\n".join(lines) + "\n")

    expanded_value = _expand_with_current_env(var_value)
    os.environ[var_name] = expanded_value

# %% ../nbs/01_env.ipynb 64
@patch
def read_user_env_var(self: WordslabEnv, var_name: str) -> Optional[str]:
    """
    Read a variable value from
    $WORDSLAB_WORKSPACE/.secrets/user-env.bashrc

    Returns the raw (unexpanded) value, or None if not found.
    """

    if not re.fullmatch(r"[A-Za-z_][A-Za-z0-9_]*", var_name):
        raise ValueError(f"Invalid environment variable name: {var_name}")

    workspace = os.environ.get("WORDSLAB_WORKSPACE")
    if not workspace:
        raise EnvironmentError("WORDSLAB_WORKSPACE is not set")

    env_file = Path(workspace) / ".secrets" / "user-env.bashrc"
    if not env_file.exists():
        return None

    pattern = re.compile(
        rf"""
        ^\s*
        (?:export\s+)?
        {re.escape(var_name)}
        \s*=\s*
        (?:
            "(?P<dq>(?:\\.|[^"])*)" |
            (?P<bare>[^\s#]+)
        )
        """,
        re.VERBOSE,
    )

    for line in env_file.read_text().splitlines():
        if line.strip().startswith("#"):
            continue

        match = pattern.match(line)
        if match:
            if match.group("dq") is not None:
                return bytes(match.group("dq"), "utf-8").decode("unicode_escape")
            return match.group("bare")

    return None

# %% ../nbs/01_env.ipynb 71
@patch
def setup_openrouter(self: WordslabEnv, openrouter_api_key: str) -> None:
    self.write_user_env_var(self.cloud_openrouter_api_key_var, openrouter_api_key)

# %% ../nbs/01_env.ipynb 75
@patch
def setup_replicate(self: WordslabEnv, replicate_api_token: str) -> None:
    self.write_user_env_var(self.cloud_replicate_api_token_var, replicate_api_token)

# %% ../nbs/01_env.ipynb 79
@patch
def setup_tavily(self: WordslabEnv, tavily_api_key: str) -> None:
    self.write_user_env_var(self.cloud_tavily_api_key_var, tavily_api_key)

# %% ../nbs/01_env.ipynb 83
@patch
def setup_ollamacloud(self: WordslabEnv, ollama_api_key: str) -> None:
    self.write_user_env_var(self.cloud_ollama_api_key_var, ollama_api_key)

# %% ../nbs/01_env.ipynb 87
@patch
def setup_github(self: WordslabEnv, git_user_name: str, git_user_email: str, github_username: str, github_access_token: str) -> None:
    # Set global Git user name and email
    subprocess.run(["git", "config", "--global", "user.name", git_user_name], check=True)
    subprocess.run(["git", "config", "--global", "user.email", git_user_email], check=True)
    # Enable Git credential store
    subprocess.run(["git", "config", "--global", "credential.helper", "store"], check=True)
    # Store credentials in ~/.git-credentials
    credentials_file = Path.home() / ".git-credentials"
    credential_line = f"https://{github_username}:{github_access_token}@github.com\n"
    with open(credentials_file, "w") as f:
        f.write(credential_line)
    print("Wrote Github credentials");

# %% ../nbs/01_env.ipynb 90
@patch
def setup_huggingface(self: WordslabEnv, huggingface_access_token: str) -> None:
    self.write_user_env_var(self.cloud_huggingface_access_token_var, huggingface_access_token)

# %% ../nbs/01_env.ipynb 94
@patch
def setup_pypi(self: WordslabEnv, pypi_api_token: str) -> None:
    pypirc_path = Path.home() / ".pypirc"
    content = f"""[pypi]
username = __token__
password = {pypi_api_token}
"""
    pypirc_path.write_text(content)
    print("Wrote Pypi credentials");
